# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from statistics import mean

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import coalesce as spark_coalesce
from pyspark.sql.functions import col as spark_col
from pyspark.sql.functions import length as spark_length
from pyspark.sql.functions import lit as spark_lit
from pyspark.sql.functions import to_binary as spark_to_binary
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_IS_UNIQUE_SIZE_KEY,
    COLUMN_MAX_SIZE_KEY,
    COLUMN_MEAN_SIZE_KEY,
    COLUMN_MIN_SIZE_KEY,
    COLUMN_SIZE_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class BinaryColumnCollector(ColumnCollectorBase):

    """Class for collect a binary type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        column_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_df: SparkDataFrame
    ) -> None:
        """Init BinaryColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

        """
        super().__init__(clm_name, struct_field, clm_df)
        self._binary_size_collection = self._compute_binary_size_collection()

    def get_custom_data(self) -> dict[str, any]:
        max_size_value = max(self._binary_size_collection)
        min_size_value = min(self._binary_size_collection)
        mean_size_value = mean(self._binary_size_collection)
        all_elements_have_same_size = max_size_value == min_size_value

        custom_data_dict = {
            COLUMN_MAX_SIZE_KEY: max_size_value,
            COLUMN_MIN_SIZE_KEY: min_size_value,
            COLUMN_MEAN_SIZE_KEY: mean_size_value,
            COLUMN_IS_UNIQUE_SIZE_KEY: all_elements_have_same_size,
        }

        return custom_data_dict

    def _compute_binary_size_collection(self) -> list[int]:
        select_result = self.column_df.select(
            spark_length(
                spark_coalesce(spark_col(self.name), spark_to_binary(spark_lit(b"")))
            ).alias(COLUMN_SIZE_KEY)
        ).collect()

        size_collection = [row[COLUMN_SIZE_KEY] for row in select_result]

        return size_collection



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Wrapper around pandera which logs to snowflake
from typing import Any, Optional, Union

from pandas import DataFrame as PandasDataFrame
from pandera import Check, DataFrameSchema
from pandera_report import DataFrameValidator

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.checkpoint_logger import CheckpointLogger
from snowflake.snowpark_checkpoints.utils.constants import (
    FAIL_STATUS,
    PASS_STATUS,
    CheckpointMode,
)
from snowflake.snowpark_checkpoints.utils.extra_config import is_checkpoint_enabled
from snowflake.snowpark_checkpoints.utils.pandera_check_manager import (
    PanderaCheckManager,
)
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _check_compare_data,
    _generate_schema,
    _process_sampling,
    _replace_special_characters,
    _update_validation_result,
)


def validate_dataframe_checkpoint(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    mode: Optional[CheckpointMode] = CheckpointMode.SCHEMA,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a Snowpark DataFrame against a specified checkpoint.

    Args:
        df (SnowparkDataFrame): The DataFrame to validate.
        checkpoint_name (str): The name of the checkpoint to validate against.
        job_context (SnowparkJobContext, optional): The job context for the validation. Required for PARQUET mode.
        mode (CheckpointMode): The mode of validation (e.g., SCHEMA, PARQUET). Defaults to SCHEMA.
        custom_checks (Optional[dict[Any, Any]], optional): Custom checks to apply during validation.
        skip_checks (Optional[dict[Any, Any]], optional): Checks to skip during validation.
        sample_frac (Optional[float], optional): Fraction of the DataFrame to sample for validation. Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample for validation.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy to use for sampling.
            Defaults to RANDOM_SAMPLE.
        output_path (Optional[str], optional): The output path for the validation results.

    Returns:
        Union[tuple[bool, PandasDataFrame], None]: A tuple containing a boolean indicating success
        and a Pandas DataFrame with validation results, or None if validation is not applicable.

    Raises:
        ValueError: If an invalid validation mode is provided or if job_context is None for PARQUET mode.

    """
    checkpoint_name = _replace_special_characters(checkpoint_name)

    if is_checkpoint_enabled(checkpoint_name):

        if mode == CheckpointMode.SCHEMA:
            return _check_dataframe_schema_file(
                df,
                checkpoint_name,
                job_context,
                custom_checks,
                skip_checks,
                sample_frac,
                sample_number,
                sampling_strategy,
                output_path,
            )
        elif mode == CheckpointMode.DATAFRAME:
            if job_context is None:
                raise ValueError(
                    "Connectionless mode is not supported for Parquet validation"
                )
            _check_compare_data(df, job_context, checkpoint_name, output_path)
        else:
            raise ValueError(
                """Invalid validation mode.
                Please use for schema validation use a 1 or for a full data validation use a 2 for schema validation."""
            )


def _check_dataframe_schema_file(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:
    """Generate and checks the schema for a given DataFrame based on a checkpoint name.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        checkpoint_name (str): The name of the checkpoint to retrieve the schema.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        tuple[bool, PanderaDataFrame]: A tuple containing the validity flag and the Pandera DataFrame.

    """
    if df is None:
        raise ValueError("DataFrame is required")

    if checkpoint_name is None:
        raise ValueError("Checkpoint name is required")

    schema = _generate_schema(checkpoint_name, output_path)

    return check_dataframe_schema(
        df,
        schema,
        checkpoint_name,
        job_context,
        custom_checks,
        skip_checks,
        sample_frac,
        sample_number,
        sampling_strategy,
        output_path,
    )


def check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a DataFrame against a given Pandera schema using sampling techniques.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (str, optional): The name of the checkpoint to retrieve the schema.
            Defaults to None.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        Union[tuple[bool, PandasDataFrame]|None]: A tuple containing the validity flag and the Pandas DataFrame.
        If the validation for that checkpoint is disabled it returns None.

    """
    checkpoint_name = _replace_special_characters(checkpoint_name)

    if df is None:
        raise ValueError("DataFrame is required")

    if pandera_schema is None:
        raise ValueError("Schema is required")

    if is_checkpoint_enabled(checkpoint_name):
        return _check_dataframe_schema(
            df,
            pandera_schema,
            checkpoint_name,
            job_context,
            custom_checks,
            skip_checks,
            sample_frac,
            sample_number,
            sampling_strategy,
            output_path,
        )


@report_telemetry(
    params_list=["pandera_schema"],
    return_indexes=[(STATUS_KEY, 0)],
    multiple_return=True,
)
def _check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: SnowparkJobContext = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:

    pandera_check_manager = PanderaCheckManager(checkpoint_name, pandera_schema)
    pandera_check_manager.skip_checks_on_schema(skip_checks)
    pandera_check_manager.add_custom_checks(custom_checks)

    pandera_schema_upper, sample_df = _process_sampling(
        df, pandera_schema, job_context, sample_frac, sample_number, sampling_strategy
    )

    # Raises SchemaError on validation issues
    validator = DataFrameValidator()
    is_valid, validation_result = validator.validate(
        pandera_schema_upper, sample_df, validity_flag=True
    )
    if is_valid:
        if job_context is not None:
            job_context._mark_pass(checkpoint_name)
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
    else:
        _update_validation_result(checkpoint_name, FAIL_STATUS, output_path)
        raise SchemaValidationError(
            "Snowpark DataFrame schema validation error",
            job_context,
            checkpoint_name,
            validation_result,
        )

    return (is_valid, validation_result)


@report_telemetry(params_list=["pandera_schema"])
def check_output_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate to validate the schema of the output of a Snowpark function.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.

    """

    def check_output_with_decorator(snowpark_fn):
        """Decorate to validate the schema of the output of a Snowpark function.

        Args:
            snowpark_fn (function): The Snowpark function to validate.

        Returns:
            function: The decorated function.

        """
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _checkpoint_name = _replace_special_characters(_checkpoint_name)

        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the output of a Snowpark function.

            Args:
                *args: The arguments to the Snowpark function.
                **kwargs: The keyword arguments to the Snowpark function.

            Returns:
                Any: The result of the Snowpark function.

            """
            # Run the sampled data in snowpark
            snowpark_results = snowpark_fn(*args, **kwargs)
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args([snowpark_results])
            pandas_sample_args = sampler.get_sampled_pandas_args()

            # Raises SchemaError on validation issues
            validator = DataFrameValidator()
            is_valid, validation_result = validator.validate(
                pandera_schema, pandas_sample_args[0], validity_flag=True
            )
            logger = CheckpointLogger().get_logger()
            logger.info(
                f"Checkpoint {_checkpoint_name} validation result:\n{validation_result}"
            )

            if is_valid:
                if job_context is not None:
                    job_context._mark_pass(_checkpoint_name)

                _update_validation_result(_checkpoint_name, PASS_STATUS, output_path)
            else:
                _update_validation_result(_checkpoint_name, FAIL_STATUS, output_path)
                raise SchemaValidationError(
                    "Snowpark output schema validation error",
                    job_context,
                    _checkpoint_name,
                    validation_result,
                )

            return snowpark_results

        return wrapper

    return check_output_with_decorator


@report_telemetry(params_list=["pandera_schema"])
def check_input_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate factory for validating input DataFrame schemas before function execution.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.


    """

    def check_input_with_decorator(snowpark_fn):
        """Decorate that validates input schemas for the decorated function.

        Args:
            snowpark_fn (Callable): The function to be decorated with input schema validation.

        Raises:
            SchemaValidationError: If input data fails schema validation.

        Returns:
            Callable: A wrapper function that performs schema validation before executing the original function.

        """
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _checkpoint_name = _replace_special_characters(_checkpoint_name)

        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the input of a Snowpark function.

            Raises:
                SchemaValidationError: If any input DataFrame fails schema validation.

            Returns:
                Any: The result of the original function after input validation.

            """
            # Run the sampled data in snowpark
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args(args)
            pandas_sample_args = sampler.get_sampled_pandas_args()

            # Raises SchemaError on validation issues
            for arg in pandas_sample_args:
                if isinstance(arg, PandasDataFrame):

                    validator = DataFrameValidator()
                    is_valid, validation_result = validator.validate(
                        pandera_schema,
                        arg,
                        validity_flag=True,
                    )

                    logger = CheckpointLogger().get_logger()
                    logger.info(
                        f"Checkpoint {checkpoint_name} validation result:\n{validation_result}"
                    )

                    if is_valid:
                        if job_context is not None:
                            job_context._mark_pass(
                                _checkpoint_name,
                            )

                        _update_validation_result(
                            _checkpoint_name,
                            PASS_STATUS,
                            output_path,
                        )
                    else:
                        _update_validation_result(
                            _checkpoint_name,
                            FAIL_STATUS,
                            output_path,
                        )
                        raise SchemaValidationError(
                            "Snowpark input schema validation error",
                            job_context,
                            _checkpoint_name,
                            validation_result,
                        )
            return snowpark_fn(*args, **kwargs)

        return wrapper

    return check_input_with_decorator



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from snowflake.snowpark_checkpoints.utils.constants import (
    BINARY_TYPE,
    BOOLEAN_TYPE,
    BYTE_TYPE,
    DATE_TYPE,
    DECIMAL_TYPE,
    DOUBLE_TYPE,
    FLOAT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    SHORT_TYPE,
    STRING_TYPE,
    TIMESTAMP_NTZ_TYPE,
    TIMESTAMP_TYPE,
)


NumericTypes = [
    BYTE_TYPE,
    SHORT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    FLOAT_TYPE,
    DOUBLE_TYPE,
    DECIMAL_TYPE,
]

StringTypes = [STRING_TYPE]

BinaryTypes = [BINARY_TYPE]

BooleanTypes = [BOOLEAN_TYPE]

DateTypes = [DATE_TYPE, TIMESTAMP_TYPE, TIMESTAMP_NTZ_TYPE]

SupportedTypes = [
    BYTE_TYPE,
    SHORT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    FLOAT_TYPE,
    DOUBLE_TYPE,
    DECIMAL_TYPE,
    STRING_TYPE,
    BINARY_TYPE,
    BOOLEAN_TYPE,
    DATE_TYPE,
    TIMESTAMP_TYPE,
    TIMESTAMP_NTZ_TYPE,
]



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from typing import Optional

from snowflake.snowpark_checkpoints.singleton import Singleton
from snowflake.snowpark_checkpoints.utils.constants import (
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    VALIDATION_RESULTS_JSON_FILE_NAME,
)
from snowflake.snowpark_checkpoints.validation_results import (
    ValidationResult,
    ValidationResults,
)


class ValidationResultsMetadata(metaclass=Singleton):

    """ValidationResultsMetadata is a class that manages the loading, storing, and updating of validation results.

    Attributes:
        validation_results (list): A list to store validation results.
        validation_results_file (str): The path to the validation results file.

    Methods:
        __init__(path: Optional[str] = None):
            Initializes the PipelineResultMetadata instance and loads validation results from a JSON file
            if a path is provided.
        _load(path: Optional[str] = None):
            Loads validation results from a JSON file. If no path is provided, the current working directory is used.
        add_validation_result(validation_result: dict):
            Adds a validation result to the pipeline result list.
        save():
            Saves the validation results to a JSON file in the current working directory.

    """

    def __init__(self, path: Optional[str] = None):
        self._load(path)

    def _load(self, path: Optional[str] = None):
        """Load validation results from a JSON file.

        Args:
            path (Optional[str]): The directory path where the validation results file is located.
                                  If not provided, the current working directory is used.

        Raises:
            Exception: If there is an error reading the validation results file.

        """
        self.validation_results_directory = path if path else os.getcwd()
        self.validation_results_directory = os.path.join(
            self.validation_results_directory,
            SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
        )

        self.validation_results_file = os.path.join(
            self.validation_results_directory,
            VALIDATION_RESULTS_JSON_FILE_NAME,
        )

        self.validation_results = ValidationResults(results=[])

        if os.path.exists(self.validation_results_file):
            with open(self.validation_results_file) as file:
                try:
                    validation_result_json = file.read()
                    self.validation_results = ValidationResults.model_validate_json(
                        validation_result_json
                    )
                except Exception as e:
                    raise Exception(
                        f"Error reading validation results file: {self.validation_results_file} \n {e}"
                    ) from None

    def clean(self):
        """Clean the validation results list.

        This method empties the validation results list.

        """
        if not os.path.exists(self.validation_results_file):
            self.validation_results.results = []

    def add_validation_result(self, validation_result: ValidationResult):
        """Add a validation result to the pipeline result list.

        Args:
            checkpoint_name (str): The name of the checkpoint.
            validation_result (dict): The validation result to be added.

        """
        self.validation_results.results.append(validation_result)

    def save(self):
        """Save the validation results to a file.

        This method checks if the directory specified by validation results directory
        exists, and if not, it creates the directory. Then, it writes the validation results
        to a file specified by validation results file in JSON format.

        Raises:
            OSError: If the directory cannot be created or the file cannot be written.

        """
        if not os.path.exists(self.validation_results_directory):
            os.makedirs(self.validation_results_directory)

        with open(self.validation_results_file, "w") as output_file:
            output_file.write(self.validation_results.model_dump_json())



from datetime import datetime
from typing import Optional

from pandera import Check, DataFrameSchema

from snowflake.snowpark_checkpoints.utils.checkpoint_logger import CheckpointLogger
from snowflake.snowpark_checkpoints.utils.constants import (
    COLUMNS_KEY,
    DECIMAL_PRECISION_KEY,
    DEFAULT_DATE_FORMAT,
    FALSE_COUNT_KEY,
    FORMAT_KEY,
    MARGIN_ERROR_KEY,
    MAX_KEY,
    MEAN_KEY,
    MIN_KEY,
    NAME_KEY,
    NULL_COUNT_KEY,
    NULLABLE_KEY,
    ROWS_COUNT_KEY,
    SKIP_ALL,
    TRUE_COUNT_KEY,
    TYPE_KEY,
)
from snowflake.snowpark_checkpoints.utils.supported_types import (
    BooleanTypes,
    NumericTypes,
)


class PanderaCheckManager:
    def __init__(self, checkpoint_name: str, schema: DataFrameSchema):
        self.checkpoint_name = checkpoint_name
        self.schema = schema

    def _add_numeric_checks(self, col: str, additional_check: dict[str, any]):
        """Add numeric checks to a specified column in the schema.

        This method adds two types of checks to the specified column:
        1. A mean check that ensures the mean of the column values is within the specified margin of error.
        2. A decimal precision check that ensures the number of decimal places in the column values does not
           exceed the specified precision.

        Args:
            col (str): The name of the column to which the checks will be added.
            additional_check (dict[str, any]): A dictionary containing the following keys:
                - MEAN_KEY: The expected mean value for the column.
                - MARGIN_ERROR_KEY: The acceptable margin of error for the mean check.
                - DECIMAL_PRECISION_KEY: The maximum number of decimal places allowed for the column values.

        """
        mean = additional_check.get(MEAN_KEY, 0)
        std = additional_check.get(MARGIN_ERROR_KEY, 0)

        def check_mean(series):
            series_mean = series.mean()
            return mean - std <= series_mean <= mean + std

        self.schema.columns[col].checks.append(
            Check(check_mean, element_wise=False, name="mean")
        )

        if DECIMAL_PRECISION_KEY in additional_check:
            self.schema.columns[col].checks.append(
                Check(
                    lambda series: series.apply(
                        lambda x: len(str(x).split(".")[1]) if "." in str(x) else 0
                    )
                    <= additional_check[DECIMAL_PRECISION_KEY],
                    name="decimal_precision",
                )
            )

    def _add_boolean_checks(self, col: str, additional_check: dict[str, any]):
        """Add boolean checks to the schema for a specified column.

        This method extends the checks for a given column in the schema by adding
        boolean checks based on the provided additional_check dictionary. It calculates
        the percentage of True and False values in the column and ensures that these
        percentages fall within a specified margin of error.

        Args:
            col (str): The name of the column to which the checks will be added.
            additional_check (dict[str, any]): A dictionary containing the following keys:
                - TRUE_COUNT_KEY: The count of True values in the column.
                - FALSE_COUNT_KEY: The count of False values in the column.
                - ROWS_COUNT_KEY: The total number of rows in the column.
                - MARGIN_ERROR_KEY: The acceptable margin of error for the percentage checks.

        Returns:
            None

        """
        count_of_true = additional_check.get(TRUE_COUNT_KEY, 0)
        count_of_false = additional_check.get(FALSE_COUNT_KEY, 0)
        rows_count = additional_check.get(ROWS_COUNT_KEY, 0)
        std = additional_check.get(MARGIN_ERROR_KEY, 0)
        percentage_true = count_of_true / rows_count
        percentage_false = count_of_false / rows_count

        self.schema.columns[col].checks.extend(
            [
                Check(
                    lambda series: (
                        percentage_true - std
                        <= series.value_counts().get(True, 0) / series.count()
                        if series.count() > 0
                        else 1 <= percentage_true + std
                    ),
                ),
                Check(
                    lambda series: (
                        percentage_false - std
                        <= series.value_counts().get(False, 0) / series.count()
                        if series.count() > 0
                        else 1 <= percentage_false + std
                    ),
                ),
            ]
        )

    def _add_null_checks(self, col: str, additional_check: dict[str, any]):
        """Add null checks to the schema for a specified column.

        This method calculates the percentage of null values in the column and
        appends a check to the schema that ensures the percentage of null values
        in the series is within an acceptable margin of error.

        Args:
            col (str): The name of the column to add null checks for.
            additional_check (dict[str, any]): A dictionary containing additional
                check parameters:
                - NULL_COUNT_KEY (str): The key for the count of null values.
                - ROWS_COUNT_KEY (str): The key for the total number of rows.
                - MARGIN_ERROR_KEY (str): The key for the margin of error.

        Raises:
            KeyError: If any of the required keys are missing from additional_check.

        """
        count_of_null = additional_check.get(NULL_COUNT_KEY, 0)
        rows_count = additional_check.get(ROWS_COUNT_KEY, 0)
        std = additional_check.get(MARGIN_ERROR_KEY, 0)
        percentage_null = count_of_null / rows_count

        self.schema.columns[col].checks.append(
            Check(
                lambda series: (
                    percentage_null - std <= series.isnull().sum() / series.count()
                    if series.count() > 0
                    else 1 <= percentage_null + std
                ),
            ),
        )

    def _add_date_time_checks(self, col: str, additional_check: dict[str, any]):
        """Add date and time checks to a specified column in the given DataFrameSchema.

        Args:
            schema (DataFrameSchema): The schema to which the checks will be added.
            col (str): The name of the column to which the checks will be applied.
            additional_check (dict[str, Any]): A dictionary containing additional check parameters.
                - FORMAT_KEY (str): The key for the date format string in the dictionary.
                - MIN_KEY (str): The key for the minimum date value in the dictionary.
                - MAX_KEY (str): The key for the maximum date value in the dictionary.

        The function will add the following checks based on the provided additional_check dictionary:
        - If both min_date and max_date are provided, a between check is added.
        - If only min_date is provided, a greater_than_or_equal_to check is added.
        - If only max_date is provided, a less_than_or_equal_to check is added.

        """
        format = additional_check.get(FORMAT_KEY, DEFAULT_DATE_FORMAT)

        min = additional_check.get(MIN_KEY, None)
        min_date = datetime.strptime(min, format) if min else None

        max = additional_check.get(MAX_KEY, None)
        max_date = datetime.strptime(max, format) if max else None

        if min_date and max_date:
            self.schema.columns[col].checks.append(
                Check.between(
                    min_date,
                    max_date,
                    include_max=True,
                    include_min=True,
                )
            )
        elif min_date:
            self.schema.columns[col].checks.append(
                Check.greater_than_or_equal_to(min_date)
            )
        elif max_date:
            self.schema.columns[col].checks.append(
                Check.less_than_or_equal_to(max_date)
            )

    def _add_date_checks(self, col: str, additional_check: dict[str, any]):
        """Add date and time checks to a specified column in the given DataFrameSchema.

        Args:
            schema (DataFrameSchema): The schema to which the checks will be added.
            col (str): The name of the column to which the checks will be applied.
            additional_check (dict[str, Any]): A dictionary containing additional check parameters.
                - FORMAT_KEY (str): The key for the date format string in the dictionary.
                - MIN_KEY (str): The key for the minimum date value in the dictionary.
                - MAX_KEY (str): The key for the maximum date value in the dictionary.

        The function will add the following checks based on the provided additional_check dictionary:
        - If both min_date and max_date are provided, a between check is added.
        - If only min_date is provided, a greater_than_or_equal_to check is added.
        - If only max_date is provided, a less_than_or_equal_to check is added.

        """
        format = additional_check.get(FORMAT_KEY, DEFAULT_DATE_FORMAT)

        min = additional_check.get(MIN_KEY, None)
        min_date = datetime.strptime(min, format).date() if min else None

        max = additional_check.get(MAX_KEY, None)
        max_date = datetime.strptime(max, format).date() if max else None

        if min_date and max_date:
            self.schema.columns[col].checks.append(
                Check.between(
                    min_date,
                    max_date,
                    include_max=True,
                    include_min=True,
                )
            )
        elif min_date:
            self.schema.columns[col].checks.append(
                Check.greater_than_or_equal_to(min_date)
            )
        elif max_date:
            self.schema.columns[col].checks.append(
                Check.less_than_or_equal_to(max_date)
            )

    def proccess_checks(self, custom_data: dict) -> DataFrameSchema:
        """Process the checks defined in the custom_data dictionary and applies them to the schema.

        Args:
            custom_data (dict): A dictionary containing the custom checks to be applied. The dictionary
                                should have a key corresponding to COLUMNS_KEY, which maps to a list of
                                column check definitions. Each column check definition should include
                                the following keys:
                                - TYPE_KEY: The type of the column (e.g., numeric, boolean, date, datetime).
                                - NAME_KEY: The name of the column.
                                - NULLABLE_KEY: A boolean indicating if the column is nullable.

        Returns:
            DataFrameSchema: The updated schema with the applied checks.

        Raises:
            ValueError: If the column name or type is not defined in the schema.

        """
        logger = CheckpointLogger().get_logger()
        for additional_check in custom_data.get(COLUMNS_KEY):

            type = additional_check.get(TYPE_KEY, None)
            name = additional_check.get(NAME_KEY, None)
            is_nullable = additional_check.get(NULLABLE_KEY, False)

            if name is None:
                raise ValueError(
                    f"Column name not defined in the schema {self.checkpoint_name}"
                )

            if type is None:
                raise ValueError(f"Type not defined for column {name}")

            if self.schema.columns.get(name) is None:
                logger.warning(f"Column {name} not found in schema")
                continue

            if type in NumericTypes:
                self._add_numeric_checks(name, additional_check)

            elif type in BooleanTypes:
                self._add_boolean_checks(name, additional_check)

            elif type == "date":
                self._add_date_checks(name, additional_check)

            elif type == "datetime":
                self._add_date_time_checks(name, additional_check)

            if is_nullable:
                self._add_null_checks(name, additional_check)

        return self.schema

    def skip_checks_on_schema(
        self,
        skip_checks: Optional[dict[str, list[str]]] = None,
    ) -> DataFrameSchema:
        """Modify the schema by skipping specified checks on columns.

        Args:
            skip_checks : Optional[dict[str, list[str]]], optional
                A dictionary where keys are column names and values are lists of check names to skip.
                If the special key 'SKIP_ALL' is present in the list of checks for a column, all checks
                for that column will be skipped. If None, no checks will be skipped.

        Returns:
            DataFrameSchema: The modified schema with specified checks skipped.

        """
        if not skip_checks:
            return self.schema

        for col, checks_to_skip in skip_checks.items():

            if col in self.schema.columns:

                if SKIP_ALL in checks_to_skip:
                    self.schema.columns[col].checks = {}
                else:
                    self.schema.columns[col].checks = [
                        check
                        for check in self.schema.columns[col].checks
                        if check.name not in checks_to_skip
                    ]

        return self.schema

    def add_custom_checks(
        self,
        custom_checks: Optional[dict[str, list[Check]]] = None,
    ):
        """Add custom checks to a Pandera DataFrameSchema.

        Args:
            schema (DataFrameSchema): The Pandera DataFrameSchema object to modify.
            custom_checks (Optional[dict[str, list[Check]]]): A dictionary where keys are column names
                                                    and values are lists of checks to add for
                                                    those columns.

        Returns:
            None

        """
        if not custom_checks:
            return self.schema

        for col, checks in custom_checks.items():

            if col in self.schema.columns:
                col_schema = self.schema.columns[col]
                col_schema.checks.extend(checks)
            else:
                raise ValueError(f"Column {col} not found in schema")

        return self.schema



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class NullColumnCollector(ColumnCollectorBase):

    """Class for collect a null type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        column_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_df: SparkDataFrame
    ) -> None:
        """Init NullColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

        """
        super().__init__(clm_name, struct_field, clm_df)

    def get_custom_data(self) -> dict[str, any]:
        return {}



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import col as spark_col
from pyspark.sql.functions import max as spark_max
from pyspark.sql.functions import mean as spark_mean
from pyspark.sql.functions import min as spark_min
from pyspark.sql.functions import stddev as spark_sdt
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_DECIMAL_PRECISION_KEY,
    COLUMN_MARGIN_ERROR_KEY,
    COLUMN_MAX_KEY,
    COLUMN_MEAN_KEY,
    COLUMN_MIN_KEY,
    INTEGER_TYPE_COLLECTION,
    get_decimal_token,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class NumericColumnCollector(ColumnCollectorBase):

    """Class for collect an empty column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        column_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.
        is_integer (boolean): a flag to indicate if values are integer or do not.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_df: SparkDataFrame
    ) -> None:
        """Init NumericColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

        """
        super().__init__(clm_name, struct_field, clm_df)
        self.is_integer = self.type in INTEGER_TYPE_COLLECTION

    def get_custom_data(self) -> dict[str, any]:
        select_result = self.column_df.select(
            spark_min(spark_col(self.name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(self.name)).alias(COLUMN_MAX_KEY),
            spark_mean(spark_col(self.name)).alias(COLUMN_MEAN_KEY),
            spark_sdt(spark_col(self.name)).alias(COLUMN_MARGIN_ERROR_KEY),
        ).collect()[0]

        min_value = select_result[COLUMN_MIN_KEY]
        max_value = select_result[COLUMN_MAX_KEY]
        mean_value = select_result[COLUMN_MEAN_KEY]
        decimal_precision = self._compute_decimal_precision()
        margin_error = select_result[COLUMN_MARGIN_ERROR_KEY]

        custom_data_dict = {
            COLUMN_MIN_KEY: min_value,
            COLUMN_MAX_KEY: max_value,
            COLUMN_MEAN_KEY: mean_value,
            COLUMN_DECIMAL_PRECISION_KEY: decimal_precision,
            COLUMN_MARGIN_ERROR_KEY: margin_error,
        }

        return custom_data_dict

    def _compute_decimal_precision(self) -> int:
        if self.is_integer:
            return 0

        decimal_part_index = 1
        decimal_token = get_decimal_token()
        max_decimal_digits_counted = 0

        row_collection = self.column_df.dropna().collect()
        for row in row_collection:
            value = row[0]
            value_str = str(value)
            value_split_by_token = value_str.split(decimal_token)
            decimal_part = value_split_by_token[decimal_part_index]
            decimal_digits_counted = len(decimal_part)
            max_decimal_digits_counted = max(
                decimal_digits_counted, max_decimal_digits_counted
            )

        return max_decimal_digits_counted



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from snowflake.snowpark_checkpoints.checkpoint import (
    check_dataframe_schema,
    check_output_schema,
    check_input_schema,
    validate_dataframe_checkpoint,
)
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.spark_migration import check_with_spark
from snowflake.snowpark_checkpoints.utils.constants import CheckpointMode

__all__ = [
    "check_with_spark",
    "SnowparkJobContext",
    "check_dataframe_schema",
    "check_output_schema",
    "check_input_schema",
    "validate_dataframe_checkpoint",
    "CheckpointMode",
]



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from statistics import mean

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import array as spark_array
from pyspark.sql.functions import coalesce as spark_coalesce
from pyspark.sql.functions import col as spark_col
from pyspark.sql.functions import explode as spark_explode
from pyspark.sql.functions import size as spark_size
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_ALLOW_NULL_KEY,
    COLUMN_IS_UNIQUE_SIZE_KEY,
    COLUMN_MAX_SIZE_KEY,
    COLUMN_MEAN_SIZE_KEY,
    COLUMN_MIN_SIZE_KEY,
    COLUMN_NULL_VALUE_PROPORTION_KEY,
    COLUMN_SIZE_KEY,
    COLUMN_VALUE_KEY,
    COLUMN_VALUE_TYPE_KEY,
    CONTAINS_NULL_KEY,
    ELEMENT_TYPE_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class ArrayColumnCollector(ColumnCollectorBase):

    """Class for collect an array type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        column_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_df: SparkDataFrame
    ) -> None:
        """Init ArrayColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

        """
        super().__init__(clm_name, struct_field, clm_df)
        self._array_size_collection = self._compute_array_size_collection()

    def get_custom_data(self) -> dict[str, any]:
        data_type_dict = self.struct_field.dataType.jsonValue()
        array_type = data_type_dict[ELEMENT_TYPE_KEY]
        allow_null = data_type_dict[CONTAINS_NULL_KEY]
        null_value_proportion = (
            self._compute_null_value_proportion() if allow_null else 0.0
        )
        array_max_size = max(self._array_size_collection)
        array_min_size = min(self._array_size_collection)
        array_mean_size = mean(self._array_size_collection)
        all_array_have_same_size = array_max_size == array_min_size

        custom_data_dict = {
            COLUMN_VALUE_TYPE_KEY: array_type,
            COLUMN_ALLOW_NULL_KEY: allow_null,
            COLUMN_NULL_VALUE_PROPORTION_KEY: null_value_proportion,
            COLUMN_MAX_SIZE_KEY: array_max_size,
            COLUMN_MIN_SIZE_KEY: array_min_size,
            COLUMN_MEAN_SIZE_KEY: array_mean_size,
            COLUMN_IS_UNIQUE_SIZE_KEY: all_array_have_same_size,
        }

        return custom_data_dict

    def _compute_array_size_collection(self) -> list[int]:
        select_result = self.column_df.select(
            spark_size(spark_coalesce(spark_col(self.name), spark_array([]))).alias(
                COLUMN_SIZE_KEY
            )
        ).collect()

        size_collection = [row[COLUMN_SIZE_KEY] for row in select_result]

        return size_collection

    def _compute_null_value_proportion(self) -> float:
        select_result = self.column_df.select(
            spark_explode(spark_col(self.name)).alias(COLUMN_VALUE_KEY)
        )

        null_counter = select_result.where(spark_col(COLUMN_VALUE_KEY).isNull()).count()

        total_values = sum(self._array_size_collection)
        null_value_proportion = (null_counter / total_values) * 100
        return null_value_proportion



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional, TypeVar

import pandas as pd

from pyspark.sql import DataFrame as SparkDataFrame

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark.types import PandasDataFrame
from snowflake.snowpark_checkpoints.errors import SparkMigrationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.constants import FAIL_STATUS, PASS_STATUS
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _replace_special_characters,
    _update_validation_result,
)


fn = TypeVar("F", bound=Callable)


def check_with_spark(
    job_context: Optional[SnowparkJobContext],
    spark_function: fn,
    checkpoint_name: str,
    sample_number: Optional[int] = 100,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Callable[[fn], fn]:
    """Validate function output with Spark instance.

    Will take the input snowpark dataframe of this function, sample data, convert
    it to a Spark dataframe and then execute `spark_function`. Subsequently
    the output of that function will be compared to the output of this function
    for the same sample of data.

    Args:
        job_context (SnowparkJobContext): The job context containing configuration and details for the validation.
        spark_function (fn): The equivalent PySpark function to compare against the Snowpark implementation.
        checkpoint_name (str): A name for the checkpoint. Defaults to None.
        sample_number (Optional[int], optional): The number of rows for validation. Defaults to 100.
        sampling_strategy (Optional[SamplingStrategy], optional): The strategy used for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (Optional[str], optional): The path to store the validation results. Defaults to None.

    Returns:
        Callable[[fn], fn]: A decorator that wraps the original Snowpark function with validation logic.

    """

    def check_with_spark_decorator(snowpark_fn):
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _checkpoint_name = _replace_special_characters(_checkpoint_name)

        def wrapper(*args, **kwargs):
            sampler = SamplingAdapter(
                job_context,
                sample_number=sample_number,
                sampling_strategy=sampling_strategy,
            )
            sampler.process_args(args)
            snowpark_sample_args = sampler.get_sampled_snowpark_args()
            pyspark_sample_args = sampler.get_sampled_spark_args()
            # Run the sampled data in snowpark
            snowpark_test_results = snowpark_fn(*snowpark_sample_args, **kwargs)
            spark_test_results = spark_function(*pyspark_sample_args, **kwargs)
            result, exception = _assert_return(
                snowpark_test_results,
                spark_test_results,
                job_context,
                _checkpoint_name,
                output_path,
            )
            if not result:
                raise exception from None
            # Run the original function in snowpark
            return snowpark_fn(*args, **kwargs)

        return wrapper

    return check_with_spark_decorator


@report_telemetry(
    params_list=["snowpark_results", "spark_results"],
    return_indexes=[(STATUS_KEY, 0)],
    multiple_return=True,
)
def _assert_return(
    snowpark_results, spark_results, job_context, checkpoint_name, output_path=None
) -> tuple[bool, Optional[Exception]]:
    """Assert and validate the results from Snowpark and Spark transformations.

    Args:
        snowpark_results (Any): Results from the Snowpark transformation.
        spark_results (Any): Results from the Spark transformation to compare against.
        job_context (Any): Additional context about the job. Defaults to None.
        checkpoint_name (Any): Name of the checkpoint for logging. Defaults to None.
        output_path (Optional[str], optional): The path to store the validation results. Defaults to None.

    Raises:
        AssertionError: If the Snowpark and Spark results do not match.
        TypeError: If the results cannot be compared.

    """
    if isinstance(snowpark_results, SnowparkDataFrame) and isinstance(
        spark_results, SparkDataFrame
    ):
        cmp = compare_spark_snowpark_dfs(spark_results, snowpark_results)

        if not cmp.empty:
            exception_result = SparkMigrationError(
                "DataFrame difference:\n", job_context, checkpoint_name, cmp
            )
            return False, exception_result
        job_context._mark_pass(checkpoint_name)
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        return True, None
    else:

        if snowpark_results != spark_results:
            exception_result = SparkMigrationError(
                "Return value difference:\n",
                job_context,
                checkpoint_name,
                f"{snowpark_results} != {spark_results}",
            )
            _update_validation_result(checkpoint_name, FAIL_STATUS, output_path)
            return False, exception_result
        job_context._mark_pass(checkpoint_name)
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        return True, None


def compare_spark_snowpark_dfs(
    spark_df: SparkDataFrame, snowpark_df: SnowparkDataFrame
) -> PandasDataFrame:
    """Compare two dataframes for equality.

    Args:
        spark_df (SparkDataFrame): The Spark dataframe to compare.
        snowpark_df (SnowparkDataFrame): The Snowpark dataframe to compare.

    Returns:Pandas DataFrame containing the differences between the two dataframes.

    """
    snowpark_df = snowpark_df.to_pandas()
    snowpark_df.columns = snowpark_df.columns.str.upper()
    spark_df = spark_df.toPandas()
    spark_df.columns = spark_df.columns.str.upper()
    spark_cols = set(spark_df.columns)
    snowpark_cols = set(snowpark_df.columns)
    cmp = pd.DataFrame([])
    left = spark_cols - snowpark_cols
    right = snowpark_cols - spark_cols
    if left != set():
        cmp = _compare_dfs(spark_df, snowpark_df, "spark", "snowpark")
    if right != set():
        right_cmp = _compare_dfs(snowpark_df, spark_df, "snowpark", "spark")
        cmp = right_cmp if cmp.empty else pd.concat([cmp, right_cmp], ignore_index=True)
    if left == set() and right == set():
        if spark_df.shape == snowpark_df.shape:
            cmp = spark_df.compare(snowpark_df, result_names=("spark", "snowpark"))
        else:
            cmp = spark_df.merge(snowpark_df, indicator=True, how="outer").loc[
                lambda x: x["_merge"] != "both"
            ]
            cmp = cmp.replace(
                {"left_only": "spark_only", "right_only": "snowpark_only"}
            )

    return cmp


def _compare_dfs(
    df_a: pd.DataFrame, df_b: pd.DataFrame, left_label: str, right_label: str
) -> PandasDataFrame:
    """Compare two dataframes for equality.

    Args:
        df_a (PandasDataFrame): The first dataframe to compare.
        df_b (PandasDataFrame): The second dataframe to compare.
        left_label (str): The label for the first dataframe.
        right_label (str): The label for the second dataframe.

    :return: Pandas DataFrame containing the differences between the two dataframes.

    """
    df_a["side"] = "a"
    df_b["side"] = "b"
    a_only = [col for col in df_a.columns if col not in df_b.columns] + ["side"]
    b_only = [col for col in df_b.columns if col not in df_a.columns] + ["side"]
    cmp = (
        df_a[a_only]
        .merge(df_b[b_only], indicator=True, how="left")
        .loc[lambda x: x["_merge"] != "both"]
    )
    cmp = cmp.replace(
        {"left_only": f"{left_label}_only", "right_only": f"{right_label}_only"}
    )
    cmp = cmp.drop(columns="side")
    return cmp



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["CollectionPointResult", "CollectionResult", "CollectionPointResultManager"]

from snowflake.snowpark_checkpoints_collector.collection_result.model.collection_point_result import (
    CollectionPointResult,
    CollectionResult,
)
from snowflake.snowpark_checkpoints_collector.collection_result.model.collection_point_result_manager import (
    CollectionPointResultManager,
)



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import col as spark_col
from pyspark.sql.functions import max as spark_max
from pyspark.sql.functions import min as spark_min
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_MAX_KEY,
    COLUMN_MIN_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class DayTimeIntervalColumnCollector(ColumnCollectorBase):

    """Class for collect a date time interval type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        column_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_df: SparkDataFrame
    ) -> None:
        """Init DayTimeIntervalColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_df (pyspark.sql.DataFrame): the column values as PySpark DataFrame.

        """
        super().__init__(clm_name, struct_field, clm_df)

    def get_custom_data(self) -> dict[str, any]:
        select_result = self.column_df.select(
            spark_min(spark_col(self.name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(self.name)).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_value = str(select_result[COLUMN_MIN_KEY])
        max_value = str(select_result[COLUMN_MAX_KEY])

        custom_data_dict = {COLUMN_MIN_KEY: min_value, COLUMN_MAX_KEY: max_value}

        return custom_data_dict
